import datetime as dt
from unittest.mock import patch

from discordproxy.exceptions import DiscordProxyTimeoutError, GrpcStatusCode
from memberaudit.tests.utils import add_memberaudit_character_to_user
from pytz import utc

from django.test import override_settings

from app_utils.testing import NoSocketsTestCase, create_fake_user

from mailrelay.tasks import (
    forward_mail_to_discord,
    forward_new_mails,
    forward_new_mails_for_config,
)

from .factories import (
    create_character_mail,
    create_eve_entities_from_evecharacter,
    create_eve_entity,
    create_relay_config,
)

MODELS_PATH = "mailrelay.models"
TASKS_PATH = "mailrelay.tasks"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 2)
@patch(TASKS_PATH + ".update_character_mails", spec=True)
@patch(TASKS_PATH + ".forward_new_mails_for_config", spec=True)
class TestForwardNewMailsAllConfigs(NoSocketsTestCase):
    def test_should_send_mails(
        self, mock_forward_new_mails_for_config, mock_update_character_mails
    ):
        # given
        user_1001 = create_fake_user(1001, "Bruce Wayne")
        character_1001 = add_memberaudit_character_to_user(user_1001, 1001)
        config_1001 = create_relay_config(character=character_1001)
        user_1002 = create_fake_user(1002, "Peter Parker")
        character_1002 = add_memberaudit_character_to_user(user_1002, 1002)
        config_1002 = create_relay_config(character=character_1002)
        user_1003 = create_fake_user(1003, "Clark Kent")
        character_1003 = add_memberaudit_character_to_user(user_1003, 1003)
        create_relay_config(character=character_1003, is_enabled=False)
        # when
        forward_new_mails()
        # then
        self.assertEqual(mock_update_character_mails.apply_async.call_count, 2)
        self.assertEqual(mock_forward_new_mails_for_config.apply_async.call_count, 2)
        called_config_pks = {
            o[1]["args"][0]
            for o in mock_forward_new_mails_for_config.apply_async.call_args_list
        }
        self.assertSetEqual({config_1001.pk, config_1002.pk}, called_config_pks)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 2)
@patch(MODELS_PATH + ".RelayConfig.send_mail")
class TestForwardNewMailsOneConfig(NoSocketsTestCase):
    def test_should_forward_all_mails(self, mock_send_mail):
        # given
        mock_send_mail.return_value = True
        user = create_fake_user(1001, "Bruce Wayne")
        character = add_memberaudit_character_to_user(user, 1001)
        create_eve_entities_from_evecharacter(character.eve_character)
        create_eve_entity(id=1002, name="Peter Parker")
        mail_1 = create_character_mail(
            character=character, sender_id=1002, recipient_ids=[2001]
        )
        mail_2 = create_character_mail(character=character, sender_id=1002)
        config = create_relay_config(character=character)
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            forward_new_mails_for_config(config_pk=config.pk)
        # then
        config.refresh_from_db()
        mails_pk = {call[1]["mail"].pk for call in mock_send_mail.call_args_list}
        self.assertSetEqual(mails_pk, {mail_1.pk, mail_2.pk})
        self.assertIsNotNone(config.last_service_run_at)

    def test_handle_no_new_mails(self, mock_send_mail):
        # given
        mock_send_mail.return_value = True
        user = create_fake_user(1001, "Bruce Wayne")
        character = add_memberaudit_character_to_user(user, 1001)
        create_eve_entities_from_evecharacter(character.eve_character)
        create_eve_entity(id=1002, name="Peter Parker")
        mail = create_character_mail(character=character, sender_id=1002)
        config = create_relay_config(character=character)
        config.mails_sent.add(mail)
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            forward_new_mails_for_config(config_pk=config.pk)
        # then
        config.refresh_from_db()
        mails_pk = {call[1]["mail"].pk for call in mock_send_mail.call_args_list}
        self.assertSetEqual(mails_pk, set())
        self.assertIsNotNone(config.last_service_run_at)


@patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 2)
@patch(MODELS_PATH + ".RelayConfig.send_mail")
class TestForwardMailToDiscord(NoSocketsTestCase):
    def test_should_send_mail(self, mock_send_mail):
        # given
        mock_send_mail.return_value = True
        user = create_fake_user(1001, "Bruce Wayne")
        character = add_memberaudit_character_to_user(user, 1001)
        create_eve_entities_from_evecharacter(character.eve_character)
        create_eve_entity(id=1002, name="Peter Parker")
        mail = create_character_mail(character=character, sender_id=1002)
        config = create_relay_config(character=character)
        # when
        forward_mail_to_discord(config_pk=config.pk, mail_pk=mail.pk)
        # then
        self.assertEqual(mock_send_mail.call_count, 1)
        _, kwargs = mock_send_mail.call_args
        self.assertEqual(kwargs["mail"], mail)

    def test_should_handle_discord_error(self, mock_send_mail):
        # given
        my_error = DiscordProxyTimeoutError(
            status=GrpcStatusCode.DEADLINE_EXCEEDED, details="test"
        )
        mock_send_mail.side_effect = my_error
        user = create_fake_user(1001, "Bruce Wayne")
        character = add_memberaudit_character_to_user(user, 1001)
        create_eve_entities_from_evecharacter(character.eve_character)
        create_eve_entity(id=1002, name="Peter Parker")
        mail = create_character_mail(character=character, sender_id=1002)
        config = create_relay_config(character=character)
        # when
        forward_mail_to_discord(config_pk=config.pk, mail_pk=mail.pk)
        # then
        pass



import datetime as dt
from unittest.mock import patch

from discordproxy.client import Channel
from memberaudit.tests.utils import add_memberaudit_character_to_user
from pytz import utc

from app_utils.testing import NoSocketsTestCase, create_fake_user

from mailrelay.models import DiscordCategory, DiscordChannel, RelayConfig

from .factories import (
    create_character_mail,
    create_discord_category,
    create_discord_channel,
    create_discordproxy_channel,
    create_eve_entities_from_evecharacter,
    create_eve_entity,
    create_relay_config,
)

MODELS_PATH = "mailrelay.models"
MANAGERS_PATH = "mailrelay.managers"


class TestRelayConfigNewMailsQueryset(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        user = create_fake_user(1001, "Bruce Wayne")
        cls.character = add_memberaudit_character_to_user(user, 1001)
        create_eve_entities_from_evecharacter(cls.character.eve_character)
        create_eve_entity(id=1002, name="Peter Parker")

    @patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 2)
    def test_should_return_corporation_mails_only(self):
        # given
        corporation_mail = create_character_mail(
            character=self.character, sender_id=1002, recipient_ids=[2001]
        )
        create_character_mail(character=self.character, sender_id=1002)
        create_character_mail(
            character=self.character, sender_id=1002, recipient_ids=[3001]
        )
        config = create_relay_config(
            character=self.character, mail_category=RelayConfig.MailCategory.CORPORATION
        )

        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            result = config.new_mails_queryset()

        # then
        mail_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(mail_pks, {corporation_mail.pk})

    @patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 2)
    def test_should_return_alliance_mails_only(self):
        # given
        alliance_mail = create_character_mail(
            character=self.character, sender_id=1002, recipient_ids=[3001]
        )
        create_character_mail(character=self.character, sender_id=1002)
        create_character_mail(
            character=self.character, sender_id=1002, recipient_ids=[2001]
        )
        config = create_relay_config(
            character=self.character, mail_category=RelayConfig.MailCategory.ALLIANCE
        )

        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            result = config.new_mails_queryset()

        # then
        mail_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(mail_pks, {alliance_mail.pk})

    @patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 2)
    def test_should_not_return_alliance_mails(self):
        # given
        user = create_fake_user(1003, "Clark Kent", 2009, "Wayne Food", "WYF")
        character = add_memberaudit_character_to_user(user, 1003)
        character.eve_character.alliance_id = None
        character.eve_character.alliance_name = ""
        character.eve_character.save()
        create_eve_entities_from_evecharacter(character.eve_character)
        create_character_mail(character=character, sender_id=1002, recipient_ids=[3001])
        create_character_mail(character=character, sender_id=1002)
        create_character_mail(character=character, sender_id=1002, recipient_ids=[2001])
        config = create_relay_config(
            character=character, mail_category=RelayConfig.MailCategory.ALLIANCE
        )

        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            result = config.new_mails_queryset()

        # then
        mail_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(mail_pks, set())

    @patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 2)
    def test_should_return_all_mails(self):
        # given
        corporation_mail = create_character_mail(
            character=self.character, sender_id=1002, recipient_ids=[2001]
        )
        personal_mail = create_character_mail(character=self.character, sender_id=1002)
        alliance_mail = create_character_mail(
            character=self.character, sender_id=1002, recipient_ids=[3001]
        )
        config = create_relay_config(
            character=self.character, mail_category=RelayConfig.MailCategory.ALL
        )

        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            result = config.new_mails_queryset()

        # then
        mail_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(
            mail_pks, {corporation_mail.pk, personal_mail.pk, alliance_mail.pk}
        )

    @patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 1)
    def test_should_not_return_old_mails(self):
        # given
        new_mail = create_character_mail(
            character=self.character,
            sender_id=1002,
            timestamp=dt.datetime(2021, 12, 24, 11, 30, tzinfo=utc),
        )
        create_character_mail(
            character=self.character,
            sender_id=1002,
            timestamp=dt.datetime(2021, 12, 24, 11, 00, tzinfo=utc),
        )
        config = create_relay_config(
            character=self.character, mail_category=RelayConfig.MailCategory.ALL
        )

        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 29, tzinfo=utc)
            result = config.new_mails_queryset()

        # then
        mail_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(mail_pks, {new_mail.pk})

    @patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 0)
    def test_should_return_all_mail_when_setting_disabled(self):
        # given
        new_mail = create_character_mail(
            character=self.character,
            sender_id=1002,
            timestamp=dt.datetime(2021, 12, 24, 11, 30, tzinfo=utc),
        )
        old_mail = create_character_mail(
            character=self.character,
            sender_id=1002,
            timestamp=dt.datetime(2021, 12, 24, 11, 00, tzinfo=utc),
        )
        config = create_relay_config(
            character=self.character, mail_category=RelayConfig.MailCategory.ALL
        )

        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 29, tzinfo=utc)
            result = config.new_mails_queryset()

        # then
        mail_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(mail_pks, {old_mail.pk, new_mail.pk})


@patch(MODELS_PATH + ".DiscordClient.create_channel_message")
class TestRelayConfigSendMail(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        user = create_fake_user(1001, "Bruce Wayne")
        cls.character = add_memberaudit_character_to_user(user, 1001)
        create_eve_entities_from_evecharacter(cls.character.eve_character)
        create_eve_entity(id=1002, name="Peter Parker")

    def test_should_send_valid_mail(self, mock_create_channel_message):
        # given
        mail = create_character_mail(character=self.character, sender_id=1002)
        config = create_relay_config(character=self.character)
        # when
        config.send_mail(mail)
        # then
        self.assertTrue(mock_create_channel_message.called)

    def test_should_not_send_mail_without_body(self, mock_create_channel_message):
        # given
        mail = create_character_mail(character=self.character, sender_id=1002, body="")
        config = create_relay_config(character=self.character)
        # when
        config.send_mail(mail)
        # then
        self.assertFalse(mock_create_channel_message.called)

    def test_should_send_mail_with_everbody_ping(self, mock_create_channel_message):
        # given
        mail = create_character_mail(character=self.character, sender_id=1002)
        config = create_relay_config(
            character=self.character, ping_type=RelayConfig.ChannelPingType.EVERYONE
        )
        # when
        config.send_mail(mail)
        # then
        _, kwargs = mock_create_channel_message.call_args
        self.assertIn("@everyone", kwargs["content"])


class TestRelayConfigOther(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        user = create_fake_user(1001, "Bruce Wayne")
        cls.character = add_memberaudit_character_to_user(user, 1001)

    def test_should_record_service_run(self):
        config = create_relay_config(character=self.character)
        my_now = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = my_now
            config.record_service_run()
        # then
        config.refresh_from_db()
        self.assertAlmostEqual(
            config.last_service_run_at, my_now, delta=dt.timedelta(seconds=30)
        )

    @patch(MODELS_PATH + ".MAILRELAY_RELAY_GRACE_MINUTES", 30)
    def test_should_report_as_up(self):
        config = create_relay_config(
            character=self.character,
            last_service_run_at=dt.datetime(2021, 12, 24, 12, 15, tzinfo=utc),
        )
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            result = config.is_service_up
        # then
        self.assertTrue(result)

    @patch(MODELS_PATH + ".MAILRELAY_RELAY_GRACE_MINUTES", 30)
    def test_should_report_as_down_1(self):
        config = create_relay_config(
            character=self.character,
            last_service_run_at=dt.datetime(2021, 12, 24, 11, 55, tzinfo=utc),
        )
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            result = config.is_service_up
        # then
        self.assertFalse(result)

    @patch(MODELS_PATH + ".MAILRELAY_RELAY_GRACE_MINUTES", 30)
    def test_should_report_as_down_2(self):
        config = create_relay_config(character=self.character, last_service_run_at=None)
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            result = config.is_service_up
        # then
        self.assertIsNone(result)


@patch(MANAGERS_PATH + ".DiscordClient.get_guild_channels", spec=True)
class TestDiscordChannelManager(NoSocketsTestCase):
    def test_should_create_new_channels_and_categories(self, mock_get_channels):
        # given
        mock_get_channels.return_value = [
            create_discordproxy_channel(id=1, name="alpha"),
            create_discordproxy_channel(id=2, name="bravo", parent_id=3),
            create_discordproxy_channel(
                id=3, name="zulu", type=Channel.Type.GUILD_CATEGORY
            ),
        ]
        # when
        result = DiscordChannel.objects.sync()
        # then
        self.assertEqual(result, 2)
        self.assertEqual(DiscordChannel.objects.count(), 2)
        obj = DiscordChannel.objects.get(id=1)
        self.assertEqual(obj.name, "alpha")
        self.assertIsNone(obj.category)
        obj = DiscordChannel.objects.get(id=2)
        self.assertEqual(obj.name, "bravo")
        self.assertEqual(obj.category.name, "zulu")

    def test_should_update_existing_channels_and_categores(self, mock_get_channels):
        # given
        mock_get_channels.return_value = [
            create_discordproxy_channel(id=1, name="alpha"),
            create_discordproxy_channel(id=2, name="bravo", parent_id=3),
            create_discordproxy_channel(
                id=3, name="zulu", type=Channel.Type.GUILD_CATEGORY
            ),
        ]
        create_discord_channel(id=1, name="update-me")
        create_discord_category(id=3, name="update-me")
        # when
        result = DiscordChannel.objects.sync()
        # then
        self.assertEqual(result, 2)
        self.assertEqual(DiscordChannel.objects.count(), 2)
        obj = DiscordChannel.objects.get(id=1)
        self.assertEqual(obj.name, "alpha")
        obj = DiscordChannel.objects.get(id=2)
        self.assertEqual(obj.name, "bravo")
        obj = DiscordCategory.objects.get(id=3)
        self.assertEqual(obj.name, "zulu")

    def test_should_remove_obsolete_channels_and_categories(self, mock_get_channels):
        # given
        mock_get_channels.return_value = [
            create_discordproxy_channel(id=1, name="alpha"),
            create_discordproxy_channel(id=2, name="bravo"),
        ]
        create_discord_channel(id=3, name="delete-me")
        create_discord_category(id=4, name="delete-me")
        # when
        result = DiscordChannel.objects.sync()
        # then
        self.assertEqual(result, 2)
        self.assertEqual(DiscordChannel.objects.count(), 2)
        obj = DiscordChannel.objects.get(id=1)
        self.assertEqual(obj.name, "alpha")
        obj = DiscordChannel.objects.get(id=2)
        self.assertEqual(obj.name, "bravo")
        self.assertFalse(DiscordCategory.objects.filter(id=4).exists())



import datetime as dt
from unittest.mock import patch

from celery import shared_task
from memberaudit.tests.utils import add_memberaudit_character_to_user
from pytz import utc

from django.test import override_settings

from app_utils.testing import NoSocketsTestCase, create_fake_user

from mailrelay.models import RelayConfig
from mailrelay.tasks import forward_new_mails

from .factories import (
    create_character_mail,
    create_eve_entities_from_evecharacter,
    create_eve_entity,
    create_relay_config,
)

MODELS_PATH = "mailrelay.models"
TASKS_PATH = "mailrelay.tasks"


@shared_task
def dummy_task(*args, **kwargs):
    """Can replace tasks that need to run."""
    pass


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODELS_PATH + ".MAILRELAY_OLDEST_MAIL_HOURS", 2)
@patch(MODELS_PATH + ".DiscordClient", spec=True)
@patch(TASKS_PATH + ".update_character_mails", new=dummy_task)
class TestForwardNewMails(NoSocketsTestCase):
    def test_should_forward_mail_with_one_config(self, mock_DiscordClient):
        # given
        user_1001 = create_fake_user(1001, "Bruce Wayne")
        character_1001 = add_memberaudit_character_to_user(user_1001, 1001)
        create_eve_entities_from_evecharacter(
            character_1001.character_ownership.character
        )
        create_eve_entity(id=1002, name="Peter Parker")
        create_character_mail(character=character_1001, sender_id=1002)
        create_relay_config(character=character_1001)

        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            forward_new_mails.delay()

        # then
        self.assertEqual(
            mock_DiscordClient.return_value.create_channel_message.call_count, 1
        )

    def test_should_forward_mail_with_multiple_configs(self, mock_DiscordClient):
        # given
        user_1001 = create_fake_user(1001, "Bruce Wayne")
        character_1001 = add_memberaudit_character_to_user(user_1001, 1001)
        create_eve_entities_from_evecharacter(
            character_1001.character_ownership.character
        )
        user_1002 = create_fake_user(1002, "Peter Parker")
        character_1002 = add_memberaudit_character_to_user(user_1002, 1002)
        create_eve_entity(id=1002, name="Peter Parker")
        create_character_mail(character=character_1001, sender_id=1002)
        create_relay_config(character=character_1001)
        create_character_mail(
            character=character_1002, sender_id=1002, recipient_ids=[2001, 1001]
        )
        create_relay_config(
            character=character_1002, mail_category=RelayConfig.MailCategory.CORPORATION
        )
        # when
        with patch(MODELS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 24, 12, 30, tzinfo=utc)
            forward_new_mails.delay()
        # then
        self.assertEqual(
            mock_DiscordClient.return_value.create_channel_message.call_count, 2
        )



"""Managers for Mail Relay."""

from typing import Optional

from discordproxy.client import Channel, DiscordClient

from django.conf import settings
from django.db import models


class DiscordChannelManager(models.Manager):
    """Manager for DiscordChannel."""

    def sync(self, timeout: Optional[int] = None) -> int:
        """Synchronize list of guild channels objects with the Discord server.

        Args:
        - timeout: timeout for request to Discord in seconds

        Returns:
        number of channels
        """
        from .models import DiscordCategory

        try:
            guild_id = int(settings.DISCORD_GUILD_ID)
        except AttributeError:
            raise ValueError(
                "Can not find Discord guild ID in settings. "
                "Is the Discord service configured?"
            ) from None
        client = DiscordClient(timeout=timeout)
        channels = client.get_guild_channels(guild_id)
        # pylint: disable=no-member
        categories = {
            obj.id: obj for obj in channels if obj.type == Channel.Type.GUILD_CATEGORY
        }
        for category in categories.values():
            DiscordCategory.objects.update_or_create(
                id=category.id, defaults={"name": category.name}
            )
        channel_ids = set()
        # pylint: disable=no-member
        text_channels = [obj for obj in channels if obj.type == Channel.Type.GUILD_TEXT]
        for channel in text_channels:
            if channel.parent_id and channel.parent_id in categories:
                category_id = channel.parent_id
            else:
                category_id = None
            self.update_or_create(
                id=channel.id,
                defaults={"name": channel.name, "category_id": category_id},
            )
            channel_ids.add(channel.id)
        self.exclude(id__in=channel_ids).delete()
        DiscordCategory.objects.filter(channels__isnull=True).delete()
        return len(text_channels)



from unittest.mock import patch

from discordproxy.exceptions import DiscordProxyException

from django.test import RequestFactory

from app_utils.testing import NoSocketsTestCase

from mailrelay.views import admin_update_discord_channels

from .factories import create_superuser

VIEWS_PATH = "mailrelay.views"


@patch(VIEWS_PATH + ".DiscordChannel.objects.sync")
@patch(VIEWS_PATH + ".messages")
class TestViews(NoSocketsTestCase):
    def test_should_post_success_message(self, mock_messages, mock_sync):
        # given
        mock_sync.return_value = 3
        factory = RequestFactory()
        request = factory.get("/mailrelay/admin_update_discord_channels/")
        request.user = create_superuser(username="Clark Kent")
        # when
        response = admin_update_discord_channels(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.success.called)

    def test_should_post_warning_message(self, mock_messages, mock_sync):
        # given
        mock_sync.side_effect = DiscordProxyException
        factory = RequestFactory()
        request = factory.get("/mailrelay/admin_update_discord_channels/")
        request.user = create_superuser(username="Clark Kent")
        # when
        response = admin_update_discord_channels(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.warning.called)



import datetime as dt

from discordproxy.discord_api_pb2 import Channel
from memberaudit.models import CharacterMail, MailEntity
from pytz import utc

from django.contrib.auth.models import User
from eveuniverse.models import EveEntity

from mailrelay.models import DiscordCategory, DiscordChannel, RelayConfig


class FakeRequest(object):
    def __init__(self, user=None):
        self.user = user


def id_generator() -> int:
    seed = 1
    while True:
        yield seed
        seed += 1


unique_ids = id_generator()


def create_eve_entity(**kwargs) -> EveEntity:
    if "category" not in kwargs:
        kwargs["category"] = EveEntity.CATEGORY_CHARACTER
    return EveEntity.objects.create(**kwargs)


def create_eve_entities_from_evecharacter(character):
    create_eve_entity(
        id=character.character_id,
        name=character.character_name,
        category=EveEntity.CATEGORY_CHARACTER,
    )
    create_eve_entity(
        id=character.corporation_id,
        name=character.corporation_name,
        category=EveEntity.CATEGORY_CORPORATION,
    )
    if character.alliance_id:
        create_eve_entity(
            id=character.alliance_id,
            name=character.alliance_name,
            category=EveEntity.CATEGORY_ALLIANCE,
        )


def create_character_mail(sender_id, recipient_ids=None, **kwargs) -> CharacterMail:
    if "timestamp" not in kwargs:
        kwargs["timestamp"] = dt.datetime(2021, 12, 24, 12, 15, tzinfo=utc)
    if not recipient_ids:
        recipient_ids = []
    if "character" not in kwargs:
        raise ValueError("character parameter not provided")
    character = kwargs["character"]
    sender, _ = MailEntity.objects.update_or_create_from_eve_entity_id(id=sender_id)
    mail_id = next(unique_ids)
    if "body" not in kwargs:
        kwargs["body"] = f"body #{mail_id}"
    kwargs.update(
        {
            "subject": f"subject #{mail_id}",
            "is_read": False,
            "mail_id": mail_id,
            "sender": sender,
        }
    )
    mail = CharacterMail.objects.create(**kwargs)
    recipient_ids += [character.eve_character.character_id]
    recipient_objs = [
        MailEntity.objects.update_or_create_from_eve_entity_id(id=recipient_id)[0]
        for recipient_id in recipient_ids
    ]
    mail.recipients.add(*recipient_objs)
    return mail


def create_relay_config(**kwargs):
    if "mail_category" not in kwargs:
        kwargs["mail_category"] = RelayConfig.MailCategory.ALL
    if "discord_channel" not in kwargs:
        kwargs["discord_channel"] = create_discord_channel(name="test")
    config = RelayConfig.objects.create(**kwargs)

    return config


def create_discord_channel(**kwargs):
    kwargs["id"] = next(unique_ids)
    return DiscordChannel.objects.create(**kwargs)


def create_discord_category(**kwargs):
    kwargs["id"] = next(unique_ids)
    return DiscordCategory.objects.create(**kwargs)


def create_superuser(**kwargs):
    return User.objects.create_superuser(**kwargs)


def create_fake_request(**kwargs):
    return FakeRequest(**kwargs)


# Discordproxy


def create_discordproxy_channel(**kwargs) -> Channel:
    if "id" not in kwargs:
        kwargs["id"] = next(unique_ids)
    if "type" not in kwargs:
        kwargs["type"] = Channel.Type.GUILD_TEXT
    return Channel(**kwargs)



from django.test import TestCase

from mailrelay.utils import chunks_by_lines


class TestChunkLines(TestCase):
    def test_should_produce_chunks(self):
        # given
        input = "abcdef\nghijklmnopq\nrstuvwxyz"
        # when
        result = chunks_by_lines(input, 20)
        # then
        self.assertListEqual(result, ["abcdef\nghijklmnopq\n", "rstuvwxyz"])



"""Tasks for Mail Relay."""

from celery import chain, shared_task
from discordproxy.exceptions import DiscordProxyException
from memberaudit.models import CharacterMail
from memberaudit.tasks import update_character_mails

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import MAILRELAY_DISCORD_TASK_TIMEOUT
from .models import RelayConfig

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

WAIT_FOR_MAIL_UPDATE_TO_COMPLETE = 60


@shared_task
def forward_new_mails():
    """Forward new mails from all active configs."""
    for config in RelayConfig.objects.filter(is_enabled=True):
        if not config.discord_channel:
            logger.warning("No channel configured for config %s", config)
            continue

        update_character_mails.apply_async(
            kwargs={"character_pk": config.character.pk, "force_update": False}
        )
        forward_new_mails_for_config.apply_async(
            args=[config.pk], countdown=WAIT_FOR_MAIL_UPDATE_TO_COMPLETE
        )


@shared_task
def forward_new_mails_for_config(config_pk: int):
    """Forward new mails from one config."""
    config = RelayConfig.objects.select_related("character").get(pk=config_pk)
    new_mails_qs = config.new_mails_queryset()
    if not new_mails_qs.exists():
        config.record_service_run()
        logger.info("No new mails to forward")
        return

    my_tasks = [
        forward_mail_to_discord.si(config_pk=config_pk, mail_pk=mail.pk)
        for mail in new_mails_qs.order_by("timestamp")
    ]
    my_tasks.append(record_service_run.si(config_pk))
    chain(my_tasks).delay()


@shared_task
def forward_mail_to_discord(config_pk, mail_pk):
    """Forward one mail to Discord."""
    config: RelayConfig = RelayConfig.objects.select_related(
        "character", "discord_channel"
    ).get(pk=config_pk)
    mail: CharacterMail = config.character.mails.get(pk=mail_pk)  # type: ignore
    try:
        config.send_mail(mail=mail, timeout=MAILRELAY_DISCORD_TASK_TIMEOUT)
    except DiscordProxyException as ex:
        logger.error(
            "%s: Failed to send mail %s to channel %s due to error from Discord Proxy. "
            "Will try again later: %s",
            config,
            mail,
            config.discord_channel,
            ex,
        )

    logger.info(
        "%s: Forwarded mail %s to channel %s", config, mail, config.discord_channel
    )


@shared_task
def record_service_run(config_pk):
    """Record completion of successful relay."""
    config = RelayConfig.objects.get(pk=config_pk)
    config.record_service_run()



# Generated by Django 4.0.10 on 2023-12-08 16:40

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mailrelay", "0002_django4_update"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="discordcategory",
            options={"verbose_name_plural": "Discord categories"},
        ),
        migrations.AlterField(
            model_name="relayconfig",
            name="ping_type",
            field=models.CharField(
                choices=[("PN", "(none)"), ("PH", "@here"), ("PE", "@everyone")],
                default="PN",
                help_text="Option to ping every member of the channel.",
                max_length=2,
                verbose_name="channel pings",
            ),
        ),
    ]



from django.test import TestCase

from mailrelay.core.xml_converter import eve_xml_to_discord_markup


class TestXmlToMarkup(TestCase):
    def test_should_replace_line_breaks(self):
        # when
        result = eve_xml_to_discord_markup("alpha<br>bravo<br><br>charlie")
        # then
        self.assertEqual(result, "alpha\nbravo\n\ncharlie")

    def test_should_convert_bold(self):
        # when
        result = eve_xml_to_discord_markup("<b>alpha</b>")
        # then
        self.assertEqual(result, "**alpha**")

    def test_should_convert_italic(self):
        # when
        result = eve_xml_to_discord_markup("<i>alpha</i>")
        # then
        self.assertEqual(result, "_alpha_")

    def test_should_convert_underline(self):
        # when
        result = eve_xml_to_discord_markup("<u>alpha</u>")
        # then
        self.assertEqual(result, "__alpha__")

    def test_should_convert_url_links(self):
        # when
        result = eve_xml_to_discord_markup(
            '<a href="https://www.example.com">alpha</a>'
        )
        # then
        self.assertEqual(result, "[alpha](https://www.example.com)")

    def test_should_handle_other_links(self):
        # when
        result = eve_xml_to_discord_markup(
            '<a href="showinfo:35825//1033237775571">alpha</a>'
        )
        # then
        self.assertEqual(result, "**alpha**")
